/*
 * ** header with license infoc
 * ...
 */

'use strict';

/**
 * @module ext_stockmanagement
 * @version 0.2
 *
 * @requires($, navbar, log, form_elements, form_panel, sample_management_reference_resolver)
 */
const ext_stockmanagement = function($, navbar, log, form_elements, form_panel, sample_management_reference_resolver) {

    const tool_box = "Storage Management" // The item that is shown in the top-navbar

    ///////////////// REGISTER NEW CONTAINERS START
    const register_new_containers_title = "Register containers";
    const register_new_containers_panel_id = "register_new_containers_form_panel";
    const container_id_label = ("${BUILD_MODULE_EXT_STOCKMANAGEMENT_ID_LABEL}" != "") ? "${BUILD_MODULE_EXT_STOCKMANAGEMENT_ID_LABEL}" : "Container ID";
    const register_new_containers_form_config = {
        // the script can be found in profiles/default/custom/caosdb-server/scripting/bin/
        // it is called like ./register_new_containers.py form.json where form.json
        // is a json dictionary cntaining a "name: value" pair for each form element.

        // NOTE: The script has to be executable ("chmod +x scriptname")
        // Else the server will resond with 400: BadRequest

        // NOTE: The she-bang (#!/usr/bin/env python3) in line 1 of the script is obligatory!
        // Else there will be syntax errors
        script: "register_new_containers.py",
        fields: [{
                type: "reference_drop_down",
                name: "responsible_entity",
                label: "Responsible entity",
                query: "FIND RECORD Responsible",
                make_desc: sample_management_reference_resolver.resolve_person_reference,
                required: true
            },
            {
                type: "reference_drop_down",
                name: "container_type",
                label: "Container type",
                query: "FIND RECORD ContainerType",
                make_desc: getEntityName,
                required: true
            },
            {
                type: "text",
                name: "container_size",
                label: "Container size",
                help: "A text describing the container size, e.g. '81 spaces' or '20*20*10 cm^3'."
            },
            {
                type: "integer",
                name: "number_of_containers",
                label: "Number of containers",
                required: true
            },
            {
                type: "select",
                name: "file_format",
                label: "Template file format",
                options: [{
                        value: "csv",
                        label: "Comma separated file (.csv)"
                    },
                    // { value: "xlsx", label: "Microsoft Excel spreadsheet (.xlsx)" }, // TODO: implement
                    // { value: "ods", label: "OpenDocument Spreadsheet (.ods)" }, // TODO: implement
                ],
                value: "csv",
                required: true
            },
            {
                type: "reference_drop_down",
                name: "parent_id",
                label: "Parent container",
                query: "FIND RECORD Container",
                make_desc: sample_management_reference_resolver.resolve_default,
                required: false
            },
        ],
    };
    ///////////////// REGISTER NEW CONTAINERS END

    ///////////////// UPDATE CONTAINERS START
    const update_containers_title = "Update Containers";
    const update_containers_panel_id = "update_containers_form_panel";
    const accepted_file_formats = [
        ".csv",
        ".CSV",
        "application/csv",
        "text/csv",
    ]
    const update_containers_form_config = {
        script: "update_containers.py",
        fields: [{
            type: "file",
            name: "container_metadata_file",
            label: "Container metadata spreadsheet",
            required: true,
            cached: false,
            accept: accepted_file_formats.join(","),
            help: "Select the filled out metadata spreadsheet you want to upload."
        }, ],
    };
    ///////////////// UPDATE CONTAINERS END

    ///////////////// EXPORT CONTAINER CSV START
    const export_csv_id = "export_container_csv";
    const export_csv_title = "Export containers to CSV";

    const export_csv_form_config = {
        script: "export_container_csv.py",
        fields: [{
            type: "text",
            name: "container_ids",
            label: container_id_label,
            required: true,
            pattern: "((([0-9]+\\s*-\\s*){0,1}[0-9]+)\\s*,\\s*)*(([0-9]+\\s*-\\s*){0,1}[0-9]+)",
            cached: false,
            help: `Comma-separated list of single ${container_id_label}s or ranges, e.g. '124, 126-139, 242'`
        }]
    };
    ///////////////// EXPORT CONTAINER CSV END


    /**
     * Add a button to the top-navbar, saying "Stock Management" which opens a
     * dropdown for Container registration and for uploading filled-in container
     * forms.
     */
    const init_show_stockmanagement_panel_buttons = function() {
        create_register_new_containers_button();
        create_update_containers_button();

        const csv_callback = form_panel.create_show_form_callback(
            export_csv_id,
            export_csv_title,
            export_csv_form_config
        );
        navbar.add_tool(export_csv_title, tool_box, {
            callback: (e) => {
                csv_callback(e);
                const input_field = document.querySelector(`#${export_csv_id} form`)[0]

                // TODO move auto focus to form_panel module
                input_field.focus();

                // TODO move pattern to form_elements module
                input_field.setAttribute("pattern", export_csv_form_config.fields[0].pattern);

                const check_pattern = () => {
                    input_field.checkValidity();
                    input_field.reportValidity();
                }
                input_field.addEventListener("blur", check_pattern);
            }
        });
    };

    const create_register_new_containers_button = function() {
        navbar.add_tool(register_new_containers_title, tool_box, {
            callback: form_panel.create_show_form_callback(
                register_new_containers_panel_id,
                register_new_containers_title,
                register_new_containers_form_config)
        });
    };

    const create_update_containers_button = function() {
        navbar.add_tool(update_containers_title, tool_box, {
            callback: form_panel.create_show_form_callback(
                update_containers_panel_id,
                update_containers_title,
                update_containers_form_config)
        });
    };

    const logger = log.getLogger("stockmanagement_form");

    /*
     * Check if the user is authenticated and has at least stockmanager or
     * administrator role.
     */
    const is_authorized = function() {
        return isAuthenticated() && (userHasRole("Scientist") || userHasRole("Stock manager") || userHasRole("administration") || userHasRole("Curator"));
    }

    /*
     * Initialization of the module. Adds the "Stock Management" Button to the
     * top-navbar is_authorized() returns true, i.e. the user is authenticated
     * and has either stockmanager or administration role.
     */
    const init = function() {
        if (is_authorized()) {
            init_show_stockmanagement_panel_buttons();
        }
    }

    /* the main function must return the initialization of the module */
    return {
        init: init,
    };
}($, navbar, log, form_elements, form_panel, sample_management_reference_resolver); // this will be replaced by require.js in the future.

$(document).ready(function() {
    // use a variable starting with `BUILD_MODULE_` to this module
    // the build variable has to be enabled in the `build.properties.d/` directory.
    // Otherwise the module will not be activated.
    if ("${BUILD_MODULE_EXT_STOCKMANAGEMENT}" === "ENABLED") {
        caosdb_modules.register(ext_stockmanagement);
    }
});
// const form = form_elements.make_form(config);
// $("body").append(form);
