#
# Copyright (C) 2025 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2025 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import re

from datetime import date
from dateutil.relativedelta import relativedelta
from typing import Union

from .utils import get_column_header_name


def semicolon_separated_list(text):
    return [el.strip() for el in text.split(";") if el != ""]


def _embargo_converter(text: str):

    datepattern = r"^(?P<year>\d{4,4})-(?P<month>\d{2,2})-(?P<day>\d{2,2})"
    matches = re.match(datepattern, str(text))
    if matches:
        return date(int(matches.groupdict()["year"]), int(matches.groupdict()["month"]), int(matches.groupdict()["day"]))
    if f"{text}".lower() in ["true", "yes"]:
        # yes means embargo until today in one year
        return date.today() + relativedelta(years=1)
    if f"{text}".lower() in ["false", "no"]:
        return ""
    raise ValueError(
        f"The embargo should be either a date in YYYY-MM-DD format, or 'true'/'yes' or 'false'/'no', but is {text}.")


def use_custom_names(definition: Union[list, dict]):
    """Replace names in list or dict keys by custom names with
    `utils.get_column_header_name`.

    """
    if isinstance(definition, list):
        return [get_column_header_name(name) for name in definition]
    elif isinstance(definition, dict):
        return {get_column_header_name(key): value for key, value in definition.items()}

    raise ValueError(f"Expected dict or list, but got {type(definition)}.")


DATATYPE_DEFINITIONS = use_custom_names({
    "entity_id": str,
    "Main User": str,
    "Campaign": str,
    "Elevation start": float,
    "Elevation stop": float,
    "End date": str,
    "Latitude start": float,
    "Latitude stop": float,
    "Longitude start": float,
    "Longitude stop": float,
    "PI": str,
    "Parent LinkAhead ID": str,
    "sample_name": str,
    "SampleMethod": str,
    "SampleType": str,
    "Start date": str,
    "Storage ID": str,
})

# Obligatory columns: Must exist and must not be empty
# Must exist
OBLIGATORY_COLUMNS = use_custom_names([
    "entity_id",
])

OBLIGATORY_COLUMNS_CHILD = use_custom_names([
    "entity_id",
    "Parent LinkAhead ID",
])

COLUMN_CONVERTER = use_custom_names({
    "Collector": semicolon_separated_list,
    "Curator": semicolon_separated_list,
    "Embargo": _embargo_converter,
    "responsible_person_event": semicolon_separated_list,
    "Sphere": semicolon_separated_list,
})

SPECIAL_TREATMENT_SAMPLE = use_custom_names([
    "Biome",
    "Campaign",
    "Collector",
    "Curator",
    "Device",
    "Elevation start",
    "Elevation stop",
    "Embargo",
    "End date",
    "responsible_person_event",
    "igsn_doi_prop",
    "Latitude start",
    "Latitude stop",
    "Level",
    "locality_description_prop",
    "locality_name_prop",
    "Longitude start",
    "Longitude stop",
    "Main User",
    "PI",
    "PDFReport",
    "parent_sample_prop",
    "sample_name",
    "Sphere",
    "Start date",
    "Storage ID",
    "entity_id",
])

IGNORED_COLUMN_NAMES_SAMPLE = use_custom_names([
    "LinkAhead URL",
    "Parent_sample",
    "Storage chain",
])
